#pragma include "utils/Preferences.frag.glsl"
#pragma include "utils/Color.glsl"
#pragma include "DefaultCommon.glsl"

// Returns the position of the point on the screen in screen pixels from 0, 0 to width, height
vec2 getScreenPositionInPixels();

// Tile texture coordinates for the sprite
vec2 tileUv(vec2 textureUv, vec2 topLeft, vec4 orientedSize);

// Returns a color from the specified texture at the specified coordinates.
// The sprite coordinates and transformation method are used. Uv coordinates are always tiled.
lowp vec4 getTextureColor(sampler2D sampler, vec2 textureUv, vec2 topLeft, vec4 orientedSize);

// Returns a color from the specified texture at the specified coordinates.
// Conversion method and tiling - standard (Z_TILED)
lowp vec4 getTextureColor(sampler2D sampler, vec2 textureUv);

// Returns the color from the standard texture at the specified coordinates.
// Conversion method and tiling - standard (Z_TILED). If there is no texture, returns white.
lowp vec4 getTextureColor(vec2 textureUv);

// Returns a color from a standard texture. Conversion method and tiling - standard (Z_TILED).
// If there is no texture, returns white.
lowp vec4 getTextureColor();

// Returns the color from the uniform. If it is not present, it returns white color.
lowp vec4 getWholeColor();

// Returns the color from the attribute. If it is not present, it returns white color.
lowp vec4 getVertexColor();

// Returns the total color from the attribute and uniforms. If it is not present, it returns white color.
lowp vec4 getVaryingColor();

// Returns the total color of the attribute, uniforms, and textures. If it is not present, it returns white color.
lowp vec4 getColor();

// Returns the total color of the attribute, uniforms, and textures at the specified coordinates. If it is not present, it returns white color.
lowp vec4 getColor(vec2 textureUv);

// Sets the color to gl_FragColor
void setColor();

// Drop transparent pixels if masked drawing mode is used.
void alphaTest();

// Run the main logic of the uber shader
void defaultMain();

// positions
varying vec2 v_position;
varying vec2 v_screenPosition;
varying vec2 v_anchorPosition;

// textures and tiles
#if defined(Z_TEXTURED)
    varying vec2 v_uv;
    uniform sampler2D u_texture;
    uniform vec2 u_textureSize;
    uniform vec2 u_texturePixelSize;

    #if defined(Z_TILED)
        uniform vec2 u_quadTopLeftUv;
        uniform vec4 u_quadUvOrientedSize;
    #endif
#endif

// colors
#if defined(Z_VERTEX_COLORED)
    varying vec4 v_color;

    #if !defined(Z_BATCHED)
        varying vec4 v_precalculatedColor;
    #endif
#endif

#if !defined(Z_BATCHED)
    uniform lowp vec4 u_color;
#endif

vec2 getScreenPositionInPixels() {
    return (v_screenPosition.xy + vec2(1.0)) / 2.0 * u_screenSize;
}

vec2 tileUv(vec2 textureUv, vec2 topLeft, vec4 orientedSize) {
    vec2 spriteUv = convertTextureUvToSprite(textureUv, topLeft, orientedSize);
    spriteUv = fract(spriteUv);
    textureUv = convertSpriteUvToTexture(spriteUv, topLeft, orientedSize);
    
    return textureUv;
}

vec2 convertAndTileUv(vec2 uv) {
    #if defined(Z_TILED)
        uv = fract(uv);
        return convertSpriteUvToTexture(uv, u_quadTopLeftUv, u_quadUvOrientedSize);
    #else
        return uv;
    #endif
}

lowp vec4 getTextureColor(sampler2D sampler, vec2 textureUv, vec2 topLeft, vec4 orientedSize) {
    return texture2D(sampler, tileUv(textureUv, topLeft, orientedSize));
}

lowp vec4 getTextureColor(sampler2D sampler, vec2 textureUv) {
    #if defined(Z_TEXTURED)
        #if defined(Z_TEXT)
            return vec4(1.0, 1.0, 1.0, texture2D(sampler, convertAndTileUv(textureUv)).r);
        #elif defined(TEXTURE_NEED_PREMULTIPY_ALPHA)
            return premultiplyAlpha(texture2D(sampler, convertAndTileUv(textureUv)));
        #else
            return texture2D(sampler, convertAndTileUv(textureUv));
        #endif
    #else
        return vec4(1, 1, 1, 1);
    #endif
}

lowp vec4 getTextureColor(vec2 textureUv) {
    #if defined(Z_TEXTURED)
        return getTextureColor(u_texture, textureUv);
    #else
        return vec4(1, 1, 1, 1);
    #endif
}

lowp vec4 getTextureColor() {
    #if defined(Z_TEXTURED)
        return getTextureColor(u_texture, v_uv);
    #else
        return vec4(1, 1, 1, 1);
    #endif
}

lowp vec4 getWholeColor() {
    #if !defined(Z_BATCHED)
        return u_color;
    #else
        return vec4(1, 1, 1, 1);
    #endif
}

lowp vec4 getVertexColor() {
    #if defined(Z_VERTEX_COLORED)
        return v_color;
    #else
        return vec4(1, 1, 1, 1);
    #endif
}

lowp vec4 getVaryingColor() {
    #if defined(Z_VERTEX_COLORED) && !defined(Z_BATCHED)
        return v_precalculatedColor;
    #elif defined(Z_VERTEX_COLORED)
        return v_color;
    #else
        return vec4(1, 1, 1, 1);
    #endif
}

lowp vec4 getColor() {
    #if defined(Z_VERTEX_COLORED)
        vec4 color = getVaryingColor() * getTextureColor();  // In this case u_color is already added to v_color
    #else
        vec4 color = getWholeColor() * getTextureColor();
    #endif

    #if defined(Z_GRAYSCALE)
        return grayscale(color);
    #else
        return color;
    #endif
}

lowp vec4 getColor(vec2 texCoord) {

    #if defined(Z_VERTEX_COLORED)
        vec4 color = getVaryingColor() * getTextureColor(texCoord);
    #else
        vec4 color = getWholeColor() * getTextureColor(texCoord);
    #endif

    #if defined(Z_GRAYSCALE)
        return grayscale(color);
    #else
        return color;
    #endif
}

void setColor() {
    gl_FragColor = getColor();
}

void alphaTest() {
    #if defined(Z_ALPHA)
        if (gl_FragColor.a < Z_ALPHA)
            discard;
    #endif
}

void defaultMain() {
    setColor();
    alphaTest();
}
